#!/usr/bin/env python
# vim:fileencoding=utf8

# Copyright (c) 2014 Florian Brucker
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Abstract syntax tree (AST) and code generation for sbl2py.
"""

import collections
import datetime
import re

import sbl2py
import sbl2py.utils


class Environment(object):
    """
    Code generation environment.

    Encapsulates the options and state of a code generation session.
    """
    def __init__(self, header=None, debug=False):
        """
        Constructor.

        ``header`` is an optional string that is inserted at the
        beginning of the generated code. It is automatically prefixed
        with comment markers.

        If ``debug`` is true then the public routines of the exported
        code return a tuple consisting of the ``_String`` wrapper and
        the ``_Program`` instance that was executed.
        """
        self.debug = debug
        self.header = header
        self.direction = 1
        self.module_code = []
        self.class_code = []
        self.init_code = []
        self.among_index = 0
        self.var_index = 0

    def claim_among_index(self):
        self.among_index += 1
        return self.among_index

    def transform_pseudo_code(self, code, tokens):
        """
        Replace placeholders in pseudo code.

        ``s`` is a string containing pseudo Python code which is
        prepared using the following steps:

        - Empty lines are removed

        - Words of the form ``<v\d*>`` are replaced with unique
          identifiers

        - Words of the form ``<t\d+>`` are replaced with the
          corresponding item in the ``tokens`` list. Indentation is
          preserved, even among multiple lines.
        """
        code = sbl2py.utils.remove_empty_lines(code)
        tokens = sbl2py.utils.extract_strings(tokens)
        for v in set(re.findall(r"<v\d*>", code)):
            unique = "var%d" % self.var_index
            self.var_index += 1
            code = code.replace(v, unique)

        def sub(match):
            return sbl2py.utils.prefix_lines(
                    tokens[int(match.group(2))], match.group(1))

        return re.sub(r"( *)<t(\d+)>", sub, code)

# Ideally we would like ``Node`` to be simply a subclass of ``list``.
# However, pyparsing does some automagic to results that are instances
# of ``list``. Since we don't want that we re-implement the desired
# functionality without subclassing ``list``.
class Node(collections.MutableSequence):
    """
    Base class for nodes in the AST.

    Note that the class hierarchy of the node classes does *not*
    represent syntactical relationships in the corresponding Snowball
    code. Each node is a list, and the AST is represented via nested
    node lists.
    """
    label = ''

    def __init__(self):
        self.children = []
        self.parent = None

    def __getitem__(self, index):
        return self.children[index]

    def __setitem__(self, index, value):
        self.children[index] = value
        value.parent = self

    def __delitem__(self, index):
        self.children[index].parent = None
        del self.children[index]

    def __len__(self):
        return len(self.children)

    def insert(self, index, value):
        self.children.insert(index, value)
        value.parent = self

    def generate_children_codes(self, env):
        """
        Generate code for each child.

        ``env`` is an instance of ``Environment``.

        Returns the code blocks generated by the children as a list.
        """
        return [node.generate_code(env) for node in self]

    def generate_code(self, env):
        """
        Generate code for this node.

        ``env`` is an instance of ``Environment``.
        """
        raise NotImplementedError('Must be implemented by subclasses.')

    def next_sibling(self):
        """
        Return next sibling node in AST.
        """
        siblings = self.parent.children
        try:
            return siblings[siblings.index(self) + 1]
        except IndexError:
            raise ValueError('Node has no next sibling.')

    def next(self):
        """
        Return next node in AST, in depth-first order.
        """
        try:
            return self.children[0]
        except IndexError:
            current = self
            while True:
                try:
                    return current.next_sibling()
                except ValueError:
                    pass
                if not current.parent:
                    return None
                current = current.parent

    def to_xml(self):
        """
        Returns an XML-like string representation of this node and its
        children.
        """
        if len(self) == 0:
            return '<%s/>' % self.__class__.__name__
        result = ['<%s>' % self.__class__.__name__]
        for child in self:
            result.append(sbl2py.utils.prefix_lines(child.to_xml(), ' '))
        result.append('</%s>' % self.__class__.__name__)
        return '\n'.join(result)

    def __repr__(self):
        content = ', '.join(repr(child) for child in self)
        return '%s(%s)' % (self.__class__.__name__, content)

    def __nonzero__(self):
        return True  # Even if no child nodes

    def annotate(self, s):
        """
        Annotate string with class label.
        """
        if not self.label:
            return s
        return sbl2py.utils.annotate(
                s, ' ' + self.label, prefix='  ', first='##', middle='#',
                last='##', single='#')

class _PseudoCodeNode(Node):
    """
    A node which generates code by transforming fixed strings of pseudo
    code.

    Subclasses should store the node's pseudo code in the ``code`` and
    ``backwards_code`` attributes, respectively. See
    ``Environment.transform_pseudo_code`` for details on the pseudo
    code format.
    """
    code = ""
    backwards_code = None

    def generate_code(self, env):
        if env.direction == 1 or self.backwards_code is None:
            code = self.code
        else:
            code = self.backwards_code
        code = env.transform_pseudo_code(code,
                self.generate_children_codes(env))
        return self.annotate(code)

class NotNode(_PseudoCodeNode):
    label = 'not'
    code = """
<v> = s.cursor
<t0>
if not r:
    s.cursor = <v>
r = not r
"""
    backwards_code = """
<v> = len(s) - s.cursor
<t0>
if not r:
    s.cursor = len(s) - <v>
r = not r
"""

class TestNode(_PseudoCodeNode):
    label = 'test'
    code = """
<v> = s.cursor
<t0>
s.cursor = <v>
"""
    backwards_code = """
<v> = len(s) - s.cursor
<t0>
s.cursor = len(s) - <v>
"""

class TryNode(_PseudoCodeNode):
    label = 'try'
    code = """
<v> = s.cursor
<t0>
if not r:
    r = True
    s.cursor = <v>
"""
    backwards_code = """
<v> = len(s) - s.cursor
<t0>
if not r:
    r = True
    s.cursor = len(s) - <v>
"""

class DoNode(_PseudoCodeNode):
    label = 'do'
    code = """
<v> = s.cursor
<t0>
s.cursor = <v>
r = True
"""
    backwards_code = """
<v> = len(s) - s.cursor
<t0>
s.cursor = len(s) - <v>
r = True
"""

class FailNode(_PseudoCodeNode):
    label = 'fail'
    code = """
<t0>
r = False
"""

class GoToNode(_PseudoCodeNode):
    label = 'goto'
    code = """
while True:
    <v> = s.cursor
    <t0>
    if r or s.cursor == s.limit:
        s.cursor = <v>
        break
    s.cursor = <v> + 1
"""
    backwards_code = """
while True:
    <v> = len(s) - s.cursor
    <t0>
    if r or s.cursor == s.limit:
        s.cursor = len(s) - <v>
        break
    s.cursor = len(s) - <v> - 1
"""

class GoPastNode(_PseudoCodeNode):
    label = 'gopast'
    code = """
while True:
    <t0>
    if r or s.cursor == s.limit:
        break
    s.cursor += 1
"""
    backwards_code = """
while True:
    <t0>
    if r or s.cursor == s.limit:
        break
    s.cursor -= 1
"""

class RepeatNode(_PseudoCodeNode):
    label = 'repeat'
    code = """
while True:
    <v> = s.cursor
    <t0>
    if not r:
        s.cursor = <v>
        break
r = True
"""
    backwards_code = """
while True:
    <v> = len(s) - s.cursor
    <t0>
    if not r:
        s.cursor = len(s) - <v>
        break
r = True
"""

class LoopNode(_PseudoCodeNode):
    label = 'loop'
    code = """
for <v> in xrange(<t0>):
    <t1>
    if not r:
        break
"""

class AtLeastNode(_PseudoCodeNode):
    label = 'atleast'
    code = """
for <v> in xrange(<t0>):
    <t1>
    if not r:
        break
if r:
    while True:
        <v> = s.cursor
        <t1>
        if not r:
            s.cursor = <v>
            break
    r = True
"""
    backwards_code = """
for <v> in xrange(<t0>):
    <t1>
    if not r:
        break
if r:
    while True:
        <v> = len(s) - s.cursor
        <t1>
        if not r:
            s.cursor = len(s) - <v>
            break
    r = True
"""

class BackwardsNode(_PseudoCodeNode):
    label = 'backwards'
    code = """
<v0> = s.cursor
<v1> = len(s) - s.limit
s.direction *= -1
s.cursor, s.limit = s.limit, s.cursor
<t0>
s.direction *= -1
s.cursor = <v0>
s.limit = len(s) - <v1>
"""

    def generate_code(self, env):
        env.direction *= -1
        code = super(BackwardsNode, self).generate_code(env)
        env.direction *= -1
        return code

def _generate_substring_code(env, index):
    code = """
a_%d = None
r = False
<v4> = s.cursor
for <v0>, <v1>, <v2> in _a_%d:
    if s.starts_with(<v0>):
        <v3> = s.cursor
        r = (not <v1>) or getattr(self, <v1>)(s)
        if r:
          s.cursor = <v3>
          a_%d = <v2>
          break
    s.cursor = <v4>
""" % (index, index, index)
    return env.transform_pseudo_code(code, [])

def _make_if_chain(blocks):
    """
    Chain code blocks together, check if ``r`` is true between them.
    """
    code = [blocks[0]]
    prefix = ''
    for block in blocks[1:]:
        code.append(prefix + 'if r:')
        prefix += '    '
        code.append(sbl2py.utils.prefix_lines(block, prefix))
    return '\n'.join(code)

class SubstringNode(Node):
    label = 'substring'
    def generate_code(self, env):
        among_index = env.claim_among_index()
        current = self.next()
        while not isinstance(current, AmongNode):
            current = current.next()
        current.among_index = among_index
        return self.annotate(_generate_substring_code(env, among_index))

class AmongNode(Node):
    label = 'among'

    def __init__(self, strings, commands, common_cmd=None):
        """
        Constructor.

        ``strings`` is a list of tuples. Each tuple contains a search
        string, its routine name (may be empty) and the string's
        command index.

        ``commands`` is the corresponding list of command nodes.

        ``common_cmd`` can be a command node and represents the extended
        ``among`` syntax::

            among ( (C1) 'foo' (C2) 'bar' (C3))

        Here, ``C1`` is the common command.
        """
        super(AmongNode, self).__init__()
        self.among_index = None
        self.strings = strings
        self.commands = commands
        self.common_cmd = common_cmd

    def generate_var(self):
        code = []
        for s in self.strings:
            routine = 'r_' + s[1] if s[1] else ''
            code.append("(%r, '%s', %d)" % (s[0], routine, s[2]))
        return '_a_%d = (' % self.among_index + ', '.join(code) + ',)'

    def generate_if_chain(self, env):
        code = []
        for index, command in enumerate(self.commands):
            if command:
                command_code = command.generate_code(env)
            else:
                command_code = 'r = True'
            command_code = sbl2py.utils.prefix_lines(command_code, '    ')
            code.append('if a_%d == %d:\n' % (self.among_index, index) +
                        command_code)
        return '\n'.join(code)

    def generate_code(self, env):
        blocks = []
        if self.among_index is None:
            self.among_index = env.claim_among_index()
            blocks.append(_generate_substring_code(env, self.among_index))
        env.module_code.append(self.generate_var())
        if self.common_cmd:
            blocks.append(self.common_cmd.generate_code(env))
        blocks.append(self.generate_if_chain(env))
        return self.annotate(_make_if_chain(blocks))

    def to_xml(self):
        cmds_xml = '\n'.join(cmd.to_xml() for cmd in self.commands if cmd)
        if cmds_xml:
            cmds_xml = sbl2py.utils.prefix_lines(cmds_xml, '  ')
            return "<AmongNode>\n%s\n</AmongNode>" % cmds_xml
        else:
            return "<AmongNode/>"

class InsertNode(_PseudoCodeNode):
    label = 'insert'
    code = """
r = s.insert(<t0>)
"""

class AttachNode(_PseudoCodeNode):
    label = 'attach'
    code = """
r = s.attach(<t0>)
"""

class ReplaceSliceNode(_PseudoCodeNode):
    label = '<-'
    code = """
r = s.set_range(self.left, self.right, <t0>)
"""

class ExportSliceNode(_PseudoCodeNode):
    label = '->'
    code = """
r = <t0>.set_chars(s.get_range(self.left, self.right))
"""

class HopNode(_PseudoCodeNode):
    label = 'hop'
    code = """
r = s.hop(<t0>)
"""

class NextNode(_PseudoCodeNode):
    label = 'next'
    code = """
r = s.hop(1)
"""

class SetLeftNode(_PseudoCodeNode):
    label = '['
    code = """
self.left = s.cursor
r = True
"""

class SetRightNode(_PseudoCodeNode):
    label = ']'
    code = """
self.right = s.cursor
r = True
"""

class SetMarkNode(_PseudoCodeNode):
    label = 'setmark'
    code = """
<t0> = s.cursor
r = True
"""

class ToMarkNode(_PseudoCodeNode):
    label = 'tomark'
    code = """
r = s.to_mark(<t0>)
"""

class AtMarkNode(_PseudoCodeNode):
    label = 'atmark'
    code = """
r = (s.cursor == <t0>)
"""

class SetNode(_PseudoCodeNode):
    label = 'set'
    code = """
<t0> = True
r = True
"""

class UnsetNode(_PseudoCodeNode):
    label = 'unset'
    code = """
<t0> = False
r = True
"""

class EmptyCommandNode(_PseudoCodeNode):
    code = """
pass
"""

class GroupingNode(_PseudoCodeNode):
    label = 'grouping check'
    code = """
if s.cursor == s.limit:
    r = False
else:
    r = s.chars[s.cursor] in <t0>
if r:
    s.cursor += 1
"""
    backwards_code = """
if s.cursor == s.limit:
    r = False
else:
    r = s.chars[s.cursor - 1] in <t0>
if r:
    s.cursor -= 1
"""

class NonNode(_PseudoCodeNode):
    label = 'negative grouping check'
    code = """
if s.cursor == s.limit:
    r = False
else:
    r = s.chars[s.cursor] not in <t0>
if r:
    s.cursor += 1
"""
    backwards_code = """
if s.cursor == s.limit:
    r = False
else:
    r = s.chars[s.cursor - 1] not in <t0>
if r:
    s.cursor -= 1
"""

class DeleteNode(_PseudoCodeNode):
    label = 'delete'
    code = """
r = s.set_range(self.left, self.right, u'')
"""

class AtLimitNode(_PseudoCodeNode):
    label = 'atlimit'
    code = """
r = (s.cursor == s.limit)
"""

class ToLimitNode(_PseudoCodeNode):
    label = 'tolimit'
    code = """
s.cursor = s.limit
r = True
"""

class StartsWithNode(_PseudoCodeNode):
    label = 'character check'
    code = """
r = s.starts_with(<t0>)
"""

class RoutineCallNode(_PseudoCodeNode):
    label = 'routine call'
    code = """
r = <t0>(s)
"""

class TrueCommandNode(_PseudoCodeNode):
    label = 'true'
    code = """
r = True
"""

class FalseCommandNode(_PseudoCodeNode):
    label = 'false'
    code = """
r = False
"""

class BooleanCommandNode(_PseudoCodeNode):
    label = 'boolean variable check'
    code = """
r = <t0>
"""

class SetLimitNode(_PseudoCodeNode):
    label = 'setlimit'
    code = """
<v0> = s.cursor
<v1> = len(s) - s.limit
<t0>
if r:
    s.limit = s.cursor
    s.cursor = <v0>
    <t1>
    s.limit = len(s) - <v1>
"""
    backwards_code = """
<v0> = len(s) - s.cursor
<v1> = s.limit
<t0>
if r:
    s.limit = s.cursor
    s.cursor = len(s) - <v0>
    <t1>
    s.limit = <v1>
"""

class BackwardModeNode(Node):
    def generate_code(self, env):
        env.direction *= -1
        self.generate_children_codes(env)
        env.direction *= -1
        return ''

class RoutineDefinitionNode(Node):

    def __init__(self, name):
        super(RoutineDefinitionNode, self).__init__()
        self.name = name

    def generate_code(self, env):
        code = ['def r_%s(self, s):' % self.name, '    r = True']
        code.append(
                sbl2py.utils.prefix_lines(self[0].generate_code(env), '    '))
        code.append('    return r')
        env.class_code.append('\n'.join(code))
        return ''

class CharSetNode(Node):

    def __init__(self, chars):
        super(CharSetNode, self).__init__()
        self.chars = chars

    def generate_code(self, env):
        return "set(%s)" % repr(self.chars)

class SetUnionNode(_PseudoCodeNode):
    code = '(<t0> | <t1>)'

class SetDifferenceNode(_PseudoCodeNode):
    code = '(<t0> - <t1>)'

class GroupingDefinitionNode(Node):
    def __init__(self, name):
        super(GroupingDefinitionNode, self).__init__()
        self.name = name

    def generate_code(self, env):
        code = '_g_%s = %s' % (self.name, self[0].generate_code(env))
        env.module_code.append(code)
        return ''

class ConcatenationNode(Node):
    def generate_code(self, env):
        return _make_if_chain(self.generate_children_codes(env))

class _IfChainNode(Node):
    not_str = ''
    def generate_code(self, env):
        if env.direction == 1:
            lines = ['<v> = s.cursor']
        else:
            lines = ['<v> = len(s) - s.cursor']
        lines.append('<t0>')
        prefix = ''
        for t in range(1, len(self)):
            lines.append(prefix + 'if ' + self.not_str +'r:')
            prefix += '    '
            if env.direction == 1:
                lines.append(prefix + 's.cursor = <v>')
            else:
                lines.append(prefix + 's.cursor = len(s) - <v>')
            lines.append(prefix + '<t%d>' % t)
        code = '\n'.join(lines)
        code = env.transform_pseudo_code(code,
                self.generate_children_codes(env))
        return self.annotate(code)

class AndNode(_IfChainNode):
    label = 'and'

class OrNode(_IfChainNode):
    label = 'or'
    not_str = 'not '

class _ReferenceNode(Node):
    prefix = ''
    suffix = ''

    def __init__(self, name):
        super(_ReferenceNode, self).__init__()
        self.name = name

    def generate_code(self, env):
        return self.prefix + self.name + self.suffix

class StringReferenceNode(_ReferenceNode):
    prefix = 'self.s_'

class CharsReferenceNode(_ReferenceNode):
    prefix = 'self.s_'
    suffix = '.chars'

class IntegerReferenceNode(_ReferenceNode):
    prefix = 'self.i_'

class BooleanReferenceNode(_ReferenceNode):
    prefix = 'self.b_'

class RoutineReferenceNode(_ReferenceNode):
    prefix = 'self.r_'

class GroupingReferenceNode(_ReferenceNode):
    prefix = '_g_'

class StringLiteralNode(Node):

    def __init__(self, string):
        super(StringLiteralNode, self).__init__()
        self.string = string

    def generate_code(self, env):
        c = repr(self.string)
        if not c.startswith('u'):
            c = 'u' + c
        return c

class IntegerLiteralNode(Node):

    def __init__(self, integer):
        super(IntegerLiteralNode, self).__init__()
        self.integer = integer

    def generate_code(self, env):
        return str(self.integer)

class MaxIntNode(_PseudoCodeNode):
    code = 'sys.maxint'

class MinIntNode(_PseudoCodeNode):
    code = '(-sys.maxint - 1)'

class CursorNode(_PseudoCodeNode):
    code = 's.cursor'

class LimitNode(_PseudoCodeNode):
    code = 's.limit'

class SizeNode(_PseudoCodeNode):
    code = 'len(s)'

class SizeOfNode(_PseudoCodeNode):
    code = 'len(<t0>)'

class _ArithmeticOperationNode(Node):
    operator = ''
    use_brackets = False

    def generate_code(self, env):
        delim = ' ' + self.operator + ' '
        code = delim.join(self.generate_children_codes(env))
        if self.use_brackets:
            code = '(' + code + ')'
        return code

class MultiplicationNode(_ArithmeticOperationNode):
    operator = '*'

class DivisionNode(_ArithmeticOperationNode):
    operator = '/'

class AdditionNode(_ArithmeticOperationNode):
    operator = '+'
    use_brackets = True

class SubtractionNode(_ArithmeticOperationNode):
    operator = '-'
    use_brackets = True

class NegationNode(_PseudoCodeNode):
    code = '(-<t0>)'

class IntegerAssignNode(_PseudoCodeNode):
    code = """
<t0> = <t1>
r = True
"""

class IntegerIncrementByNode(_PseudoCodeNode):
    label = '+='
    code = """
<t0> += <t1>
r = True
"""

class IntegerMultiplyByNode(_PseudoCodeNode):
    label = '*='
    code = """
<t0> *= <t1>
r = True
"""

class IntegerDecrementByNode(_PseudoCodeNode):
    label = '-='
    code = """
<t0> -= <t1>
r = True
"""

class IntegerDivideByNode(_PseudoCodeNode):
    label = '/='
    code = """
<t0> /= <t1>
r = True
"""

class IntegerEqualNode(_PseudoCodeNode):
    label = '=='
    code = """
r = <t0> == <t1>
"""

class IntegerGreaterNode(_PseudoCodeNode):
    label = '>'
    code = """
r = <t0> > <t1>
"""

class IntegerLessNode(_PseudoCodeNode):
    label = '<'
    code = """
r = <t0> < <t1>
"""

class IntegerUnequalNode(_PseudoCodeNode):
    label = '!='
    code = """
r = <t0> != <t1>
"""

class IntegerGreaterOrEqualNode(_PseudoCodeNode):
    label = '>='
    code = """
r = <t0> >= <t1>
"""

class IntegerLessOrEqualNode(_PseudoCodeNode):
    label = '<='
    code = """
r = <t0> <= <t1>
"""

class _InitDeclarationNode(Node):
    def __init__(self, name):
        super(_InitDeclarationNode, self).__init__()
        self.name = name

    def generate_code(self, env):
        env.init_code.append(self.template % self.name)
        return ''

class IntegerDeclarationNode(_InitDeclarationNode):
    template = 'self.i_%s = 0'

class StringDeclarationNode(_InitDeclarationNode):
    template = "self.s_%s = _String('')"

class BooleanDeclarationNode(_InitDeclarationNode):
    template = 'self.b_%s = True'

class BooleanDeclarationNode(_InitDeclarationNode):
    template = 'self.b_%s = True'

_FUNC_TEMPLATE = """
def %s(s):
    s = _String(s)
    _Program().r_%s(s)
    return unicode(s)
"""

_DEBUG_FUNC_TEMPLATE = """
def %s(s):
    p = _Program()
    s = _String(s)
    p.r_%s(s)
    return s, p
"""

class ExternalDeclarationNode(Node):
    def __init__(self, name):
        super(ExternalDeclarationNode, self).__init__()
        self.name = name

    def generate_code(self, env):
        template = _FUNC_TEMPLATE if not env.debug else _DEBUG_FUNC_TEMPLATE
        env.module_code.append(template % (self.name, self.name))
        return ''

class _NoOpDeclarationNode(Node):
    def __init__(self, name):
        super(_NoOpDeclarationNode, self).__init__()
        self.name = ''

    def generate_code(self, env):
        return ''

class RoutineDeclarationNode(_NoOpDeclarationNode):
    pass

class GroupingDeclarationNode(_NoOpDeclarationNode):
    pass

_MODULE_TEMPLATE = """#!/usr/bin/env python
# vim:fileencoding=utf-8

%(header)s

# This file was auto-generated on %(date)s using
# version %(version)s of the sbl2py Snowball-to-Python compiler.

import sys

class _String(object):

    def __init__(self, s):
        self.chars = list(unicode(s))
        self.cursor = 0
        self.limit = len(s)
        self.direction = 1

    def __unicode__(self):
        return u''.join(self.chars)

    def __len__(self):
        return len(self.chars)

    def get_range(self, start, stop):
        if self.direction == 1:
            return self.chars[start:stop]
        else:
            n = len(self.chars)
            return self.chars[stop:start]

    def set_range(self, start, stop, chars):
        if self.direction == 1:
            self.chars[start:stop] = chars
        else:
            self.chars[stop:start] = chars
        change = self.direction * (len(chars) - (stop - start))
        if self.direction == 1:
            if self.cursor >= stop:
                self.cursor += change
                self.limit += change
        else:
            if self.cursor > start:
                self.cursor += change
            if self.limit > start:
                self.limit += change
        return True

    def insert(self, chars):
        self.chars[self.cursor:self.cursor] = chars
        if self.direction == 1:
            self.cursor += len(chars)
            self.limit += len(chars)
        return True

    def attach(self, chars):
        self.chars[self.cursor:self.cursor] = chars
        if self.direction == 1:
            self.limit += len(chars)
        else:
            self.cursor += len(chars)
        return True

    def set_chars(self, chars):
        self.chars = chars
        if self.direction == 1:
            self.cursor = 0
            self.limit = len(chars)
        else:
            self.cursor = len(chars)
            self.limit = 0
        return True

    def starts_with(self, chars):
        n = len(chars)
        r = self.get_range(self.cursor, self.limit)[::self.direction][:n]
        if not r == list(chars)[::self.direction]:
            return False
        self.cursor += n * self.direction
        return True

    def hop(self, n):
        if n < 0 or len(self.get_range(self.cursor, self.limit)) < n:
            return False
        self.cursor += n * self.direction
        return True

    def to_mark(self, mark):
        if self.direction == 1:
            if self.cursor > mark or self.limit < mark:
                return False
        else:
            if self.cursor < mark or self.limit > mark:
                return False
        self.cursor = mark
        return True

%(module_code)s

class _Program(object):
    def __init__(self):
        self.left = None
        self.right = None
%(init_code)s

%(class_code)s
"""

class ProgramNode(Node):
    """
    Root node representing a complete Snowball program.
    """
    def generate_code(self, env):
        self.generate_children_codes(env)
        if env.header:
            header = sbl2py.utils.prefix_lines(env.header, '# ')
        else:
            header = ''
        date = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        version = sbl2py.__version__
        module_code = '\n'.join(env.module_code)
        init_code = sbl2py.utils.prefix_lines(
                '\n'.join(env.init_code), '        ')
        class_code = sbl2py.utils.prefix_lines(
                '\n\n'.join(env.class_code), '    ')
        return _MODULE_TEMPLATE % {
            'header':header,
            'date':date,
            'version':version,
            'module_code':module_code,
            'init_code':init_code,
            'class_code':class_code,
        }



#!/usr/bin/env python
# vim:fileencoding=utf8

# Copyright (c) 2014 Florian Brucker
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A Snowball to Python compiler.
"""

import sbl2py.ast
import sbl2py.grammar


__version__ = '0.1.0-2'


def translate_file(infile, *args, **kwargs):
    """
    Translate a Snowball file to Python.

    ``infile`` is an open readable file containing the Snowball source
    code. The return value is a string containing the translated Python
    code.

    See ``translate_string`` for additional arguments.
    """
    return translate_string(infile.read(), *args, **kwargs)

def translate_string(code, header=None, debug=False):
    """
    Translate a Snowball code string to Python.

    ``header`` is an optional string that is inserted at the beginning
    of the generated code. It is automatically prefixed with comment
    markers.

    If ``debug`` is ``True`` then the external Snowball routines return
    both the original ``_String`` object and the ``_Program`` instance
    that created it. This is useful for checking that variables have
    been computed correctly.
    """
    node = sbl2py.grammar.parse_string(code)
    env = sbl2py.ast.Environment(header=header, debug=debug)
    return node.generate_code(env)




#!/usr/bin/env python
# vim:fileencoding=utf8

# Copyright (c) 2014 Florian Brucker
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import argparse

from sbl2py import translate_file

def main():
    parser = argparse.ArgumentParser(description='Compile Snowball to Python')
    parser.add_argument('infile', help='Input file (default STDIN)', nargs='?',
            type=argparse.FileType('r'), default=sys.stdin)
    parser.add_argument('outfile', help='Output file (default STDOUT)', nargs='?',
            type=argparse.FileType('w'), default=sys.stdout)
    parser.add_argument('-d', '--debug', help='Generate code for easier debugging',
            action="store_true")
    args = parser.parse_args()
    args.outfile.write(translate_file(args.infile, debug=args.debug) + "\n")

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# vim:fileencoding=utf8

# Copyright (c) 2014 Florian Brucker
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Snowball grammar and parser for sbl2py.
"""

import functools
import inspect
import sys
import threading

from pyparsing import *

from sbl2py.ast import *


__all__ = ['parse_string']


# Grammar elements are in all-caps.


ParserElement.enablePackrat()


#
# PARSER STATE
#

state = threading.local()
state.strings = []        # Declared string names
state.integers = []       # Declared integer names
state.externals = []      # Declared externals names
state.booleans = []       # Declared boolean names
state.routines = []       # Declared routine names
state.groupings = []      # Declared grouping names
state.stringescapes = []  # Left and right string escape chars
state.stringdefs = {}     # String replacement definitions

def reset():
    """
    Reset internal parser state.
    """
    state.strings[:] = []
    state.integers[:] = []
    state.externals[:] = []
    state.booleans[:] = []
    state.routines[:] = []
    state.groupings[:] = []
    state.stringescapes[:] = []
    state.stringdefs.clear()


#
# UTILITY FUNCTIONS
#

def parse_action(f):
    """
    Decorator for pyparsing parse actions to ease debugging.

    pyparsing uses trial & error to deduce the number of arguments a
    parse action accepts. Unfortunately any ``TypeError`` raised by a
    parse action confuses that mechanism.

    This decorator replaces the trial & error mechanism with one based
    on reflection. If the decorated function itself raises a
    ``TypeError`` then that exception is re-raised if the wrapper is
    called with less arguments than required. This makes sure that the
    actual ``TypeError`` bubbles up from the call to the parse action
    (instead of the one caused by pyparsing's trial & error).
    """
    num_args = len(inspect.getargspec(f).args)
    if num_args > 3:
        raise ValueError('Input function must take at most 3 parameters.')

    @functools.wraps(f)
    def action(*args):
        if len(args) < num_args:
            if action.exc_info:
                raise (action.exc_info[0], action.exc_info[1],
                       action.exc_info[2])
        action.exc_info = None
        try:
            v = f(*args[:-(num_args + 1):-1])
            return v
        except TypeError as e:
            action.exc_info = sys.exc_info()
            raise

    action.exc = None
    return action


def make_node_action(cls, ungroup=False, init_args=0):
    @parse_action
    def action(tokens):
        if ungroup:
            tokens = tokens[0]
        node = cls(*tokens[:init_args])
        node.extend(tokens[init_args:])
        return node
    return action


def add_node_action(pattern, *args, **kwargs):
    pattern.addParseAction(make_node_action(*args, **kwargs))
    return pattern


def make_binary_op_list_action(operators, classes, ungroup=False):
    """
    Make parse action for lists of binary operators and operands.

    ``operators`` is a list of the operators (as strings) and
    ``classes`` is a list of the corresponding ``Node`` subclasses.
    """
    @parse_action
    def action(tokens):
        if ungroup:
            tokens = tokens[0]
        tokens = list(reversed(tokens))
        left = tokens.pop()
        while tokens:
            token = tokens.pop()
            for op, cls in zip(operators, classes):
                if token == op:
                    node = cls()
                    break
            node.append(left)
            node.append(tokens.pop())
            left = node
        return node

    return action

LPAREN = Suppress('(')
RPAREN = Suppress(')')


#
# KEYWORDS
#

keywords = []

def make_keyword(s):
    kw = Keyword(s)
    globals()[s.upper()] = kw
    keywords.append(kw)

map(make_keyword, """maxint minint cursor limit size sizeof or and strings
integers booleans routines externals groupings define as not test try do fail
goto gopast repeat loop atleast insert attach delete hop next setmark tomark
atmark tolimit atlimit setlimit for backwards reverse substring among set unset
non true false backwardmode stringescapes stringdef hex decimal""".split())

KEYWORD = MatchFirst(keywords)


#
# NAMES
#

NAME = ~KEYWORD + Word(alphas, alphanums + '_')
NAME.setParseAction(parse_action(lambda t: t[0]))


#
# DECLARATIONS
#

def make_decl(kw, targets, cls):
    declaration = Suppress(kw) + LPAREN + ZeroOrMore(NAME) + RPAREN

    @parse_action
    def action(tokens):
        for target in targets:
            target.extend(tokens)
        return [cls(t) for t in tokens]

    declaration.setParseAction(action)
    return declaration

DECLARATION = MatchFirst([
    make_decl(STRINGS, [state.strings], StringDeclarationNode),
    make_decl(INTEGERS, [state.integers], IntegerDeclarationNode),
    make_decl(BOOLEANS, [state.booleans], BooleanDeclarationNode),
    make_decl(ROUTINES, [state.routines], RoutineDeclarationNode),
    make_decl(EXTERNALS, [state.externals, state.routines],
              ExternalDeclarationNode),
    make_decl(GROUPINGS, [state.groupings], GroupingDeclarationNode),
])


#
# REFERENCES
#

reference_chars = set(alphanums + '_')

class Reference(Token):
    """
    A reference to a previously declared variable.

    This class works like pyparsing's ``Or`` in combination with
    ``Keyword``. However, the list of candidates can be updated later
    on.
    """

    def __init__(self, declarations):
        """
        Constructor.

        ``declarations`` is a list of previously declared variables.
        Any of them will match if they occur as a separate word (cf.
        ``Keyword``). Matching is done in decreasing length of
        candidates (cf. ``Or``). Later updates of ``declarations``
        are taken into account.
        """
        super(Reference, self).__init__()
        self.declarations = declarations

    def __str__(self):
        return 'Reference(%s)' % self.declarations

    def parseImpl(self, instring, loc, doActions=True):
        candidates = sorted(self.declarations, key=lambda x: len(x),
                            reverse=True)
        for candidate in candidates:
            if instring.startswith(candidate, loc):
                n = len(candidate)
                if (len(instring) == loc + n or instring[loc + n] not in
                        reference_chars):
                    return loc + n, candidate
        raise ParseException("Expected one of " + ", ".join(candidates))


def make_reference(declarations, cls):
    pattern = Reference(declarations)
    pattern.setParseAction(parse_action(lambda t: cls(t[0])))
    return pattern

STR_REF = make_reference(state.strings, StringReferenceNode)
CHARS_REF = make_reference(state.strings, CharsReferenceNode)
GROUPING_REF = make_reference(state.groupings, GroupingReferenceNode)
INT_REF = make_reference(state.integers, IntegerReferenceNode)
BOOLEAN_REF = make_reference(state.booleans, BooleanReferenceNode)
ROUTINE_REF = make_reference(state.routines, RoutineReferenceNode)


#
# STRINGS
#

class StringLiteral(Token):
    """
    String literal that supports dynamically changing escape characters.
    """

    def __init__(self, escape_chars, replacements):
        """
        Constructor.

        ``escape_chars`` is a list containing either two or no
        characters. These characters are the left and right escape
        marker, respectively. You may change the content of the list
        afterwards, the parsing code always uses the latest values.

        ``replacements`` is a dict that maps escape sequences to their
        replacements. Later modifications are taken into account.
        """
        super(StringLiteral, self).__init__()
        self.escape_chars = escape_chars
        self.replacements = replacements

    def __str__(self):
        if self.escape_chars:
            return 'StringLiteral("%s%s")' % tuple(self.escape_chars)
        else:
            return 'StringLiteral("")'

    def parseImpl(self, instring, loc, doActions=True):
        if instring[loc] != "'":
            raise ParseException('Expected "\'".')
        # Find next "'" that is not contained in escape chars
        pos = loc + 1
        while True:
            try:
                candidate = instring.index("'", pos)
            except ValueError:
                raise ParseException('Runaway string literal.')
            if not self.escape_chars:
                break
            left = instring.rfind(self.escape_chars[0], loc, candidate)
            right = instring.rfind(self.escape_chars[1], loc, candidate)
            if right >= left:
                break
            pos = candidate + 1
        s = instring[loc + 1 : candidate]
        if self.escape_chars:
            # Replace escape sequences
            left = re.escape(self.escape_chars[0])
            right = re.escape(self.escape_chars[1])
            for k, v in self.replacements.iteritems():
                s = re.sub(left + re.escape(k) + right, v, s)
        return candidate + 1, unicode(s)



@parse_action
def stringescapes_cmd_action(tokens):
    state.stringescapes[:] = tokens[0] + tokens[1]
    state.stringdefs["'"] = "'"
    state.stringdefs['['] = '['
    return []

@parse_action
def stringdef_cmd_action(tokens):
    key = tokens[0]
    mode = tokens[1]
    value = tokens[2].string
    if mode == 'hex':
        value = u''.join(unichr(int(x, 16)) for x in value.split())
    elif mode == 'decimal':
        value = u''.join(unichr(int(x)) for x in value.split())
    state.stringdefs[key] = value
    return []

STR_LITERAL = StringLiteral(state.stringescapes, state.stringdefs)
STR_LITERAL.setParseAction(parse_action(lambda t: StringLiteralNode(t[0])))

CHAR = Word(printables, exact=1)
STRINGESCAPES_CMD = Suppress(STRINGESCAPES) + CHAR + CHAR
STRINGESCAPES_CMD.setParseAction(stringescapes_cmd_action)

STRINGDEF_CMD = (Suppress(STRINGDEF) + Word(printables) +
                 Optional(HEX | DECIMAL, default=None) + STR_LITERAL)
STRINGDEF_CMD.setParseAction(stringdef_cmd_action)

# A sequence of characters
CHARS = STR_LITERAL | CHARS_REF


#
# INTEGERS
#

@parse_action
def int_literal_action(tokens):
    return IntegerLiteralNode(int(tokens[0]))

INT_LITERAL = Word(nums)
INT_LITERAL.setParseAction(int_literal_action)

INT = INT_REF | INT_LITERAL


#
# ARITHMETIC EXPRESSIONS
#

EXPRESSION_OPERAND = MatchFirst([
    add_node_action(Suppress(MAXINT), MaxIntNode),
    add_node_action(Suppress(MININT), MinIntNode),
    add_node_action(Suppress(CURSOR), CursorNode),
    add_node_action(Suppress(LIMIT), LimitNode),
    add_node_action(Suppress(SIZEOF) + STR_REF, SizeOfNode),
    add_node_action(Suppress(SIZE), SizeNode),
    INT,
])

negation_action =  make_node_action(NegationNode, ungroup=True)
multiplicative_action = make_binary_op_list_action(
        ['*', '/'], [MultiplicationNode, DivisionNode])
additive_action = make_binary_op_list_action(
        ['+', '-'], [AdditionNode, SubtractionNode])

EXPRESSION = operatorPrecedence(
    EXPRESSION_OPERAND,
    [
        (Suppress('-'), 1, opAssoc.RIGHT,negation_action),
        (oneOf('* /'), 2, opAssoc.LEFT, multiplicative_action),
        (oneOf('+ -'), 2, opAssoc.LEFT, additive_action),
    ]
)


#
# INTEGER COMMANDS
#

def make_int_cmd(op, cls):
    VAR = Suppress('$') + INT_REF
    # We're not using ``Combine`` here because ``Combine`` automatically
    # converts the result to a string.
    VAR.leaveWhitespace()
    return add_node_action(VAR + Suppress(op) + EXPRESSION, cls)

INT_CMD = MatchFirst([
    make_int_cmd('+=', IntegerIncrementByNode),
    make_int_cmd('*=', IntegerMultiplyByNode),
    make_int_cmd('-=', IntegerDecrementByNode),
    make_int_cmd('/=', IntegerDivideByNode),
    make_int_cmd('==', IntegerEqualNode),
    make_int_cmd('!=', IntegerUnequalNode),
    make_int_cmd('>=', IntegerGreaterOrEqualNode),
    make_int_cmd('<=', IntegerLessOrEqualNode),
    make_int_cmd('=', IntegerAssignNode),
    make_int_cmd('>', IntegerGreaterNode),
    make_int_cmd('<', IntegerLessNode),
])


#
# STRING COMMANDS
#


STR_CMD = Forward()

UNARY_OPERATOR = (NOT | TEST | TRY | DO | FAIL | GOTO | GOPAST | REPEAT |
                  BACKWARDS | (LOOP + EXPRESSION) | (ATLEAST + EXPRESSION))

not_action = make_node_action(NotNode)
test_action = make_node_action(TestNode)
try_action = make_node_action(TryNode)
do_action = make_node_action(DoNode)
fail_action = make_node_action(FailNode)
goto_action = make_node_action(GoToNode)
gopast_action = make_node_action(GoPastNode)
repeat_action = make_node_action(RepeatNode)
loop_action = make_node_action(LoopNode)
atleast_action = make_node_action(AtLeastNode)
backwards_action = make_node_action(BackwardsNode)

unary_actions = {
    'not':not_action,
    'test':test_action,
    'try':try_action,
    'do':do_action,
    'fail':fail_action,
    'goto':goto_action,
    'gopast':gopast_action,
    'repeat':repeat_action,
    'loop':loop_action,
    'atleast':atleast_action,
    'backwards':backwards_action,
}

@parse_action
def unary_action(tokens):
    return unary_actions[tokens[0][0]](tokens[0][1:])

CMD_INSERT = add_node_action(Suppress(INSERT | '<+') + CHARS, InsertNode)
CMD_ATTACH = add_node_action(Suppress(ATTACH) + CHARS, AttachNode)
CMD_REPLACE_SLICE = add_node_action(Suppress('<-') + CHARS, ReplaceSliceNode)
CMD_EXPORT_SLICE = add_node_action(Suppress('->') + STR_REF, ExportSliceNode)
CMD_HOP = add_node_action(Suppress(HOP) + EXPRESSION, HopNode)
CMD_NEXT = add_node_action(Suppress(NEXT), NextNode)
CMD_SET_LEFT = add_node_action(Suppress('['), SetLeftNode)
CMD_SET_RIGHT = add_node_action(Suppress(']'), SetRightNode)
CMD_SETMARK = add_node_action(Suppress(SETMARK) + INT_REF, SetMarkNode)
CMD_TOMARK = add_node_action(Suppress(TOMARK) + EXPRESSION, ToMarkNode)
CMD_ATMARK = add_node_action(Suppress(ATMARK) + EXPRESSION, AtMarkNode)
CMD_SET = add_node_action(Suppress(SET) + BOOLEAN_REF, SetNode)
CMD_UNSET = add_node_action(Suppress(UNSET) + BOOLEAN_REF, UnsetNode)
CMD_GROUPING = add_node_action(GROUPING_REF.copy(), GroupingNode)
CMD_NON = add_node_action(Suppress(NON + Optional('-')) + GROUPING_REF, NonNode)
CMD_DELETE = add_node_action(Suppress(DELETE), DeleteNode)
CMD_ATLIMIT = add_node_action(Suppress(ATLIMIT), AtLimitNode)
CMD_TOLIMIT = add_node_action(Suppress(TOLIMIT), ToLimitNode)
CMD_STARTSWITH = add_node_action(CHARS.copy(), StartsWithNode)
CMD_ROUTINE = add_node_action(ROUTINE_REF.copy(), RoutineCallNode)
CMD_TRUE = add_node_action(Suppress(TRUE), TrueCommandNode)
CMD_FALSE = add_node_action(Suppress(FALSE), FalseCommandNode)
CMD_BOOLEAN = add_node_action(BOOLEAN_REF.copy(), BooleanCommandNode)
CMD_SUBSTRING = add_node_action(Suppress(SUBSTRING), SubstringNode)
CMD_SETLIMIT = add_node_action(
        Suppress(SETLIMIT) + STR_CMD + Suppress(FOR) + LPAREN + STR_CMD +
        RPAREN, SetLimitNode)
CMD_EMPTY = add_node_action(LPAREN + RPAREN, EmptyCommandNode)

@parse_action
def cmd_among_action(tokens):
    common_cmd, tokens = tokens
    strings = []
    for index, arg in enumerate(tokens):
        for element in arg[0]:
            string = element[0].string
            routine = element[1].name if element[1] else ''
            strings.append((string, routine, index))
    strings.sort(cmp=lambda x, y: len(y[0]) - len(x[0])) # by decreasing length
    commands = [arg[1] for arg in tokens]
    return AmongNode(strings, commands, common_cmd=common_cmd)

AMONG_STR = Group(STR_LITERAL + Optional(ROUTINE_REF, default=None))
AMONG_CMD_ARG = Optional((LPAREN + STR_CMD + RPAREN | CMD_EMPTY), default=None)
AMONG_ARG = Group(Group(OneOrMore(AMONG_STR)) + AMONG_CMD_ARG)
CMD_AMONG = (Suppress(AMONG) + LPAREN + AMONG_CMD_ARG +
        Group(OneOrMore(AMONG_ARG)) + RPAREN)
CMD_AMONG.setParseAction(cmd_among_action)

STR_CMD_OPERAND = (INT_CMD | CMD_STARTSWITH | CMD_SETLIMIT | CMD_INSERT |
                   CMD_ATTACH | CMD_REPLACE_SLICE | CMD_DELETE | CMD_HOP |
                   CMD_NEXT | CMD_SET_LEFT | CMD_SET_RIGHT | CMD_EMPTY |
                   CMD_EXPORT_SLICE | CMD_SETMARK | CMD_TOMARK | CMD_ATMARK |
                   CMD_TOLIMIT | CMD_ATLIMIT | CMD_SET | CMD_UNSET |
                   CMD_SUBSTRING | CMD_AMONG | CMD_ROUTINE | CMD_GROUPING |
                   CMD_NON | CMD_TRUE | CMD_FALSE | CMD_BOOLEAN)

concatenation_action = make_node_action(ConcatenationNode, ungroup=True)
and_or_action = make_binary_op_list_action(
        ['and', 'or'], [AndNode, OrNode], ungroup=True)

STR_CMD << operatorPrecedence(
    STR_CMD_OPERAND,
    [
        (UNARY_OPERATOR, 1, opAssoc.RIGHT, unary_action),
        (AND | OR, 2, opAssoc.LEFT, and_or_action),
        (Empty(), 2, opAssoc.LEFT, concatenation_action),
    ]
)


#
# ROUTINES
#

@parse_action
def routine_def_action(tokens):
    node = RoutineDefinitionNode(tokens[0].name)
    node.append(tokens[1])
    return node

ROUTINE_DEF = Suppress(DEFINE) + ROUTINE_REF + Suppress(AS) + STR_CMD
ROUTINE_DEF.setParseAction(routine_def_action)


#
# GROUPINGS
#

@parse_action
def grouping_def_action(tokens):
    tokens = list(reversed(tokens))
    node = GroupingDefinitionNode(tokens.pop().name)
    if not tokens:
        return node
    elif len(tokens) == 1:
        node.append(tokens.pop())
        return node
    left = tokens.pop()
    while tokens:
        child = SetUnionNode() if tokens.pop() == '+' else SetDifferenceNode()
        child.append(left)
        child.append(tokens.pop())
        left = child
    node.append(child)
    return node

@parse_action
def char_set_action(tokens):
    return CharSetNode(tokens[0].string)

CHAR_SET = STR_LITERAL.copy().addParseAction(char_set_action)
GROUPING_ATOM = GROUPING_REF | CHAR_SET
GROUPING_DEF = (Suppress(DEFINE) + GROUPING_REF + GROUPING_ATOM +
                ZeroOrMore(oneOf('+ -') + GROUPING_ATOM))
GROUPING_DEF.setParseAction(grouping_def_action)


#
# PROGRAM
#

PROGRAM_ATOM = (DECLARATION | ROUTINE_DEF | GROUPING_DEF | STRINGESCAPES_CMD |
                STRINGDEF_CMD)
BACKWARD_SECTION = add_node_action(
        Suppress(BACKWARDMODE) + LPAREN + ZeroOrMore(PROGRAM_ATOM) + RPAREN,
        BackwardModeNode)
PROGRAM = add_node_action(
        ZeroOrMore(PROGRAM_ATOM | BACKWARD_SECTION) + StringEnd(), ProgramNode)
PROGRAM.ignore(cStyleComment | dblSlashComment)


#
# PUBLIC INTERFACE
#

def parse_string(s):
    """
    Parse string containing Snowball code.

    Returns the corresponding AST.
    """
    reset()
    return PROGRAM.parseString(s)[0]



#!/usr/bin/env python
# vim:fileencoding=utf8

# Copyright (c) 2014 Florian Brucker
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Various utilities.
"""

import imp
import itertools
import math
import re


def remove_empty_lines(s):
    """
    Remove empty lines from a string.
    """
    return '\n'.join(line for line in s.splitlines() if line)

def prefix_lines(s, p):
    """
    Prefix each line of ``s`` by ``p``.
    """
    try:
        return p + ('\n' + p).join(s.splitlines())
    except Exception as e:
        import traceback
        traceback.print_exc()
        import pdb; pdb.set_trace()
        raise

def extract(seq, f):
    """
    Extract items from a possibly nested sequence.

    All items of the possibly nested sequence ``seq`` for which the
    callback ``f`` returns ``True`` are returned in a list.
    """
    try:
        it = iter(seq)
    except TypeError:
        return []
    items = []
    for item in it:
        if f(item):
            items.append(item)
        else:
            items.extend(extract(item, f))
    return items

def extract_strings(seq):
    """
    Extract all strings from a possibly nested sequence.
    """
    return extract(seq, lambda x: isinstance(x, basestring))

def annotate(text, caption, prefix='', single='> ', first='\\ ', middle=' | ',
             last='/ '):
    """
    Annotate a block of text with a caption.

    ``text`` is the text to be annotated. ``caption`` must be a
    single-line string.

    ``prefix`` is a string that is prefixed before each added string.
    ``single``, ``first``, ``middle`` and ``last`` are the prefixes for
    the caption bracket markers: ``single`` is used for texts that
    consist of a single line. For multi-line texts, the first line is
    marked with ``first``, the middle lines with ``middle`` and the
    last one with ``last``.
    """
    lines = text.splitlines()
    lengths = [len(line) for line in lines]
    c = max(lengths)
    n = len(lines)
    for i in range(n):
        if n == 1:
            s = single
        elif i == 0:
            s = first
        elif i == n - 1:
            s = last
        else:
            s = middle
        lines[i] = lines[i] + ' ' * (c - lengths[i]) + prefix + s
    lines[n / 2] += caption
    return '\n'.join(lines)

def group(iterable, size):
    """
    Group an iterable into tuples.
    """
    it = iter(iterable)
    while True:
        sub_it = itertools.islice(it, size)
        yield tuple([sub_it.next()] + list(sub_it))

def add_line_numbers(text, margin="  "):
    """
    Add line numbers to a text.
    """
    lines = text.splitlines()
    num_digits = math.floor(math.log10(len(lines))) + 1
    format_str = '%%%dd%s%%s' % (num_digits, margin)
    nums = xrange(1, len(lines) + 1)
    return '\n'.join(format_str % (n, line) for (n, line) in zip(nums, lines))

def module_from_code(name, code):
    """
    Dynamically create Python module from code string.
    """
    if isinstance(code, unicode):
        # Remove encoding declaration if present
        lines = code.splitlines()
        for i in (0, 1):
            if re.search(r"coding[:=]\s*([-\w.]+)", lines[i]) is not None:
                lines[i] = ''
                break
        code = '\n'.join(lines)

    module = imp.new_module(name)
    exec code in module.__dict__
    return module



import codecs
import os.path
import re
import sys

from setuptools import setup, find_packages

# We want the value of ``sbl2py.__version__``. However, we cannot
# simply ``import sbl2py`` since sbl2py requires pyparsing which
# might not be installed. Hence we extract the version information
# "manually".
module_dir = os.path.dirname(__file__)
init_filename = os.path.join(module_dir, 'sbl2py', '__init__.py')
with codecs.open(init_filename, 'r' ,'utf8') as f:
    for line in f:
        m = re.match(r'\s*__version__\s*=\s*[\'"](.*)[\'"]\s*', line)
        if m:
            version = m.group(1)
            break
    else:
        raise Exception('Could not find version number.')

setup(
    name='sbl2py',
    version=version,
    description='A Snowball-to-Python compiler',
    url='https://github.com/torfuspolymorphus/sbl2py',
    author='Florian Brucker',
    author_email='mail@florianbrucker.de',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Text Processing :: Linguistic',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Operating System :: OS Independent ',
    ],
    keywords='snowball compiler',
    packages=find_packages(exclude='test'),
    install_requires=['pyparsing >= 2.0'],
    platforms=['any'],
    entry_points={'console_scripts':['sbl2py=sbl2py.__main__:main']},
)
